/* LinuxBot Server - allows control, via TCP/IP, of LinuxBot from remote client

   Modified to use closed loop library - robot.c, 20/9/00

   Copyright (c) 2000 University of the West of England, Bristol
   
   Intelligent Autonomous Systems Lab
   Faculty of Engineering
   UWE, Bristol
   
   Alan.Winfield@uwe.ac.uk

   TO COMPILE:
       gcc lserver.c -o lserver -O

   TO USE:
       enter 'lserver' (no arguments), or
       enter 'lserver -v' for verbose mode (printf received commands)
   NOTE:
       include file linuxbot.h contains robot i/o port addresses:
           you must ensure these are correct for your robot, ie
           Moorebot has different addresses to RASCAL
*/

/*-----------------------------------------------------------------------------
   REVISION HISTORY

V1.1	20/9/00		Modified to use robot.c library
V1.2	20/9/00		Modified to allow remote selection of 
			motion control modes   
V1.3	21/9/00		Modified to allow a program to be started
V1.4    4/4/01          Updated to v1.5 robot.c library
-----------------------------------------------------------------------------*/

/* Includes start here.. */
#include <stdio.h>		/* for printf etc */
#include <string.h>		/* for strcmp() */

#include <sys/types.h>		/* for sockets API... */
#include <sys/socket.h>
#include <netdb.h>
#include <errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>

#define LINUXBOT 1
#include "robot.c"		/* closed loop library */

/* Defines start here.. */
#define SERVPORT 1025		/* Port no for the tcp/ip connection */
#define BUFLEN   128		/* length of tcp/ip i/o buffers */

/* Global variables start here.. */
int	ListSock;		/* listening socket */
int	ConnSock;		/* connected socket */
int	Verbose;		/* verbose mode (for debug) */

robot   Bot;			/* robot data structure */

/* Function prototypes.. */
int ExecCmd(char *InBuf, int InLen);

/* Main program start here.. */
int main (int argc, char **argv)
{
	struct sockaddr_in LocalAddr;	/* socket address structures */
	struct sockaddr_in RemotAddr;	
	struct linger      TimeOut;	/* timeout structure */
		
	int 	AddrLen = sizeof(LocalAddr);	/* useful extras */
	int	One = 1;
	
	char	InBuf[BUFLEN];		/* input buffer */
	int	InLen;			/* bytes in input buffer */
	int	Finished = 0;		/* flag set=0 when exit received */
	
	Verbose = 0;			/* default to verbose mode off */
	if ((argc == 2) && (strcmp(argv[1], "-v") == 0 ))
		Verbose = 1;		/* set verbose mode on -v */
		 
	printf("\nLinuxBot tcp/ip Server v1.4\n");
	
	/* First get access permission to i/o ports.. */
	if ( !accessPorts() )
	{
		printf("Port access denied... root privileges needed\n");
		exit(1);
	}
	resetBot( &Bot );		/* enable motors, etc */
	setMode( &Bot, IV_MODE );	/* start in integral velocity mode */
	
	/* Get a socket for listening.. */
	ListSock = socket(AF_INET, SOCK_STREAM, 0);
	if (ListSock < 0)
	{
		perror("socket() failed");
		exit(2);
	}
		
	/* Initialise local sockaddr_in structure.. */
	LocalAddr.sin_family      = AF_INET;		/* address family */
	LocalAddr.sin_addr.s_addr = htonl(INADDR_ANY);	/* wildcard ip address */
	LocalAddr.sin_port        = htons(SERVPORT);	/* port no */
	
	/* Bind the local address & port to the socket.. */
	if (bind(ListSock, (struct sockaddr *)&LocalAddr, AddrLen) < 0)
	{
		perror("bind() failed");
		exit(4);
	}
	
	/* Passively open the socket.. */
	if (listen(ListSock, 1) < 0)
	{
		perror("listen() failed");
		exit(5);
	}
	
	/* Wait for client to connect.. */
	printf("listening...\n");
	ConnSock = accept(ListSock, (struct sockaddr *)&RemotAddr, &AddrLen);
	if (ConnSock < 0)
	{
		perror("accept() failed");
		exit(6);
	}
	printf("...connected\n");
	
	/* Now close listening socket (accepting no more connections).. */
	close(ListSock);
	
	/* Set connected socket to non-blocking mode.. */
	if (ioctl(ConnSock, FIONBIO, (char *)&One))
		perror("ioctl() non-block failed");
		
	/* Set close linger to 0 to abort on error.. */
	TimeOut.l_onoff = 1;
	TimeOut.l_linger = 0;
	if (setsockopt(ConnSock, SOL_SOCKET, SO_LINGER,
	               (char *)&TimeOut, sizeof(TimeOut)))
		perror("setsockopt() SO_LINGER failed");
		
	/* Now loop waiting for commands to be received from client.. */
	while( !Finished )
	{
		/* Do a non-blocking read.. */
		if ((InLen = read(ConnSock, InBuf, BUFLEN)) <= 0)
		{
			if (errno != EWOULDBLOCK)
			{
				perror("read() failed");
				close(ConnSock);
				exit(8);
			}
		}
		else	/* something has been received.. */
		{
			InBuf[InLen] = 0;	/* add terminator */
		/*	printf("%s\n", InBuf);	/* display it */
			/* execute received commands.. */
			if (Finished = ExecCmd(InBuf, InLen))
				close(ConnSock);			
		}
	} /* while */
	exit(0);
}


/* Execute the commands in the receive buffer 'InBuf', of length 'InLen'.
   Note that there may be more than one command in the buffer, so we
   need to be smart about parsing these commands.
   
   Returns: 1 if we've received the 'exit' command,
            0 if not.

   This routine assumes that commands are separated by at least 1 space.
*/
int ExecCmd(char *InBuf, int InLen)
{
	char	OutBuf[BUFLEN];		/* output buffer */
	int	OutLen;			/* bytes in output buffer */
	
	int	Cmd;			/* received command */
	long	Param1, Param2;		/* command parameters */
	int	Status;			/* variable for status read */
	
	int	ExitFlag = 0;		/* exit flag, 0 = not finished */	
	char 	*InPtr = InBuf;		/* pointer to start of InBuf */

	do
	{
		/* parse the command.. */
		Cmd = (int)strtol(InPtr, &InPtr, 10);
			
		switch(Cmd)
		{
		    	case 3: 	/* STOP_CMD */
				if (Verbose) printf("cmd=%d (all stop)\n", Cmd);
				velMove( &Bot, 0.0, 0.0 );
				break;

		    	case 4:		/* STATUS_CMD */
				if (Verbose) printf("cmd=%d (status)\n", Cmd);
				Status = getSwitches();
				OutLen = sprintf(OutBuf, "status=%d", Status);
				write(ConnSock, OutBuf, OutLen);
				break;  	

		    	case 6: 	/* MOTORS_CMD */
				Param1 = strtol(InPtr, &InPtr, 10);
				Param2 = strtol(InPtr, &InPtr, 10);
				if (Verbose) printf("cmd=%d (motors on), param1=%d, param2=%d\n",
				                     Cmd, (int)Param1, (int)Param2);
				velMove( &Bot, Param1/100.0, Param2/100.0 );
				break;

			case 90:	/* MODE_CMD */
				Param1 = strtol(InPtr, &InPtr, 10);
				if (Verbose) printf("cmd=%d (mode) %d\n",
				                     Cmd, (int)Param1);
				if (Param1 == 1)
					setMode( &Bot, IDLE_MODE );
				else if (Param1 == 2)
					setMode( &Bot, PV_MODE );
				else if (Param1 == 3)
					setMode( &Bot, IV_MODE );
				break;		
			
			case 99: 	/* EXIT_CMD */
				if (Verbose ) printf("cmd=%d (exit)\n", Cmd);
				velMove( &Bot, 0.0, 0.0 );
				ExitFlag = 1;
				break;
				
			case 199:	/* execute another programme */
				if ( execve("avoid", NULL, NULL) == -1 )
					printf("execve() failed\n");
				break;

			default:	/* default.. */
				if (Verbose) printf("cmd %d not recognised\n", Cmd);
		} 
	}	/* keep looping until there are less than 3 chars left.. */
	while( (InLen - (int)(InPtr - InBuf)) >= 3 );
	
	return( ExitFlag );
}
